require 'graphql'
require 'json'
require 'net/http'
require 'netrc'

module GraphQL
  class REPL
    IntrospectionDocument = GraphQL.parse(GraphQL::Introspection::INTROSPECTION_QUERY)

    def initialize(uri, username: nil, password: nil, netrc: false, headers: {})
      @uri = URI.parse(uri)
      @headers = headers
      @username, @password = Netrc.read[@uri.hostname] if netrc

      @connection = Net::HTTP.new(@uri.host, @uri.port).tap do |client|
        client.use_ssl = @uri.scheme == "https"
      end

      @schema = GraphQL::Schema::Loader.load(request(IntrospectionDocument.to_query_string))
    end

    def request(query)
      request = Net::HTTP::Post.new(@uri.request_uri)
      request.basic_auth(@username, @password)
      request["Accept"] = "application/json"
      request["Content-Type"] = "application/json"
      @headers.each { |k, v| request[k] = v }
      request.body = JSON.generate({"query" => query})
      response = @connection.request(request)
      JSON.parse(response.body)
    end

    def eval(str)
      expr = Expression.new(@schema, str)

      if error = expr.error_message
        return error
      end

      result = request(expr.document.to_query_string)

      if result["errors"] && result["errors"][0]
        result["errors"][0]["message"]
      else
        expr.response_path.reduce(result["data"]) { |obj, key| obj[key] }
      end
    end

    def completions(str)
      Expression.new(@schema, str).completions
    end

    class Proxy
      def self.const_missing(name)
        name
      end

      attr_accessor :calls

      def initialize
        @calls = []
      end

      def get_binding
        binding
      end

      def method_missing(*args)
        @calls << args
        self
      end
    end

    class Expression
      def initialize(schema, str)
        @schema = schema
        @str = str
      end

      def calls
        proxy = Proxy.new
        b = proxy.get_binding
        b.eval(@str).calls
      end

      def completions
        first, partial = @str.scan(/^(.+)\.(.*)$/)[0]

        if first
          type = Expression.new(@schema, first).return_type
        else
          partial = @str
          type = @schema.query.to_s
        end

        if type
          @schema.types[type].all_fields.map(&:name).grep(/^#{partial}/).map { |name|
            [first, name].compact.join(".")
          }
        else
          []
        end
      end

      def response_path
        calls.reduce([]) do |path, (name, *args)|
          case name
          when :query, :mutation, :on
            path
          when :[]
            path << args[0]
          else
            path << name.to_s
          end
        end
      end

      def document
        calls = self.calls

        operation_type = case calls[0][0]
        when :query
          calls.shift
          "query"
        when :mutation
          calls.shift
          "mutation"
        else
          "query"
        end

        selection = calls.reverse.reduce(nil) do |child, call|
          case call[0]
          when :[]
            child
          when :on
            type = GraphQL::Language::Nodes::TypeName.new(name: call[1].to_s)
            node = GraphQL::Language::Nodes::InlineFragment.new(type: type)
            node.selections = [child] if child
            node
          else
            arguments = []

            (call[1] || {}).each do |name, value|
              arguments << GraphQL::Language::Nodes::Argument.new(name: name.to_s, value: value)
            end

            node = GraphQL::Language::Nodes::Field.new(name: call[0].to_s, arguments: arguments)
            node.selections = [child] if child
            node
          end
        end

        operation = GraphQL::Language::Nodes::OperationDefinition.new(operation_type: operation_type, selections: [selection])
        GraphQL::Language::Nodes::Document.new(definitions: [operation])
      end

      def return_type
        visitor = GraphQL::Language::Visitor.new(document)
        type_stack = GraphQL::StaticValidation::TypeStack.new(@schema, visitor)

        stack = []

        visitor[GraphQL::Language::Nodes::OperationDefinition] << ->(node, _parent) do
          stack << type_stack.object_types.last
        end
        visitor[GraphQL::Language::Nodes::FragmentDefinition] << ->(node, _parent) do
          stack << type_stack.object_types.last
        end
        visitor[GraphQL::Language::Nodes::Field] << ->(node, _parent) do
          stack << type_stack.field_definitions.last&.type
        end
        visitor[GraphQL::Language::Nodes::InlineFragment] << ->(node, _parent) do
          stack << type_stack.object_types.last
        end
        visitor.visit

        type = stack.last
        type = type.of_type if type.is_a?(GraphQL::NonNullType)

        if type.is_a?(GraphQL::ListType)
          if response_path.last.is_a?(Integer)
            type.unwrap.name
          else
            "List"
          end
        elsif type
          type.name
        else
          nil
        end
      end

      def error_message
        validator = GraphQL::StaticValidation::Validator.new(schema: @schema)
        query = GraphQL::Query.new(@schema, document: document)
        errors = validator.validate(query)
        errors.fetch(:errors)[0]&.message
      end
    end
  end
end
